module net.BurtonRadons.spyl.valueFloat;

private import net.BurtonRadons.spyl.value;
private import net.BurtonRadons.spyl.valueBoolean;
private import net.BurtonRadons.spyl.valueFunction;
private import net.BurtonRadons.spyl.valueInteger;
private import net.BurtonRadons.spyl.valueType;
private import std.c.stdio;
private import std.math, std.math2;

class FloatType : TypeValue
{
    private static FloatType singleton;
    private static bit locked = true;

    this ()
    {
        assert (!locked);
    }

    static FloatType create ()
    {
        if (singleton === null)
        {
            locked = false;
            singleton = new FloatType;
            locked = true;
        }
        return singleton;
    }

    char [] doc ()
    {
        return "A floating-point type.  The type has the constants "
            "infinity and nan.  Float "
            "constants can be specified by such numbers as 2.4 or 2e4.";
    }

    Value call (Value [] args)
    {
        Value result;

        assert (args.length == 1);
        result = args [0].castTo (this);
        assert (cast(FloatValue) result);
        return result;
    }

    override Value getattr (char [] name)
    {
        switch (name)
        {
            case "infinity": return new AttributeFunctionValue (&infinity, "This constant is positive infinity.");
            case "nan": return new AttributeFunctionValue (&nan, "This is a not-a-number constant.");
            default: return super.getattr (name);
        }
    }

    Value epsilon () { return new FloatValue (real.epsilon); }
    Value infinity () { return new FloatValue (real.infinity); }
    Value nan () { return new FloatValue (real.nan); }
}

class FloatValue : Value
{
    real content; /**< Actual value. */

    /** Assign the parameter. */
    this (real content)
    {
        this.content = content;
    }

    override char [] typeName ()
    {
        return "Float";
    }

    override char [] repr ()
    {
        char [640] buffer;
        int length;

        length = sprintf (buffer, "%Lg", content);
        return buffer [0 .. length].dup;
    }

    override Value getattr (char [] name)
    {
        switch (name)
        {
            case "isfinite": return new AttributeFunctionValue (&isfinite, "True if the value is not infinite.");
            case "isinfinite": return new AttributeFunctionValue (&isinfinite, "True if the value is infinite.");
            case "isnan": return new AttributeFunctionValue (&isnan, "True if the value is not a number.");
            case "sqrt": return new AttributeFunctionValue (&sqrt, "The square root of the value.");
            default: return super.getattr (name);
        }
    }

    Value isfinite () { return BooleanValue.create (cast(bit) std.math.isfinite (content)); }
    Value isinfinite () { return BooleanValue.create (cast(bit) !std.math.isfinite (content)); }
    Value isnan () { return BooleanValue.create (cast(bit) std.math2.isnan (content)); }
    Value sqrt () { return new FloatValue (std.math2.sqrt (content)); }

    override Value opPow (Value other)
    {
        real exponent;

        if (cast(IntegerValue) other)
            exponent = (cast(IntegerValue) other).content;
        else if (cast(FloatValue) other)
            exponent = (cast(FloatValue) other).content;
        else
            return super.opPow (other);

        return new FloatValue (std.math2.pow (content, exponent));
    }

    override Value opPow_r (Value other)
    {
        version (none) {   /*+++*/ //ERROR: function pow overloads real(real x,uint n) and real(real x,real y) both match argument list for pow
            if (cast(IntegerValue) other)
                return new FloatValue (std.math2.pow ((cast(IntegerValue) other).content, content));
            return super.opPow_r (other);
        } else {           /*+++*/
            if (cast(IntegerValue) other)
                return new FloatValue (std.math2.pow (cast(real)(cast(IntegerValue) other).content, content));
            else if (cast(FloatValue) other)
                return new FloatValue (std.math2.pow (cast(real)(cast(FloatValue) other).content, content));
            return super.opPow_r (other);
        }
    }

    override Value castTo (Value type)
    {
        if (cast(FloatType) type)
            return this;
        if (cast(IntegerType) type)
        {
            if (!std.math.isfinite (content) || std.math2.isnan (content))
                throw new Error ("Cannot cast non-normal floats to integer.");
            return new IntegerValue (cast(long)(content + 0.5));                     // add (long) /*+++*/
        }
        return super.castTo (type);
    }
}
